program linear (input,output);

{Modification of kripke program for linear logic, from 25 March 1992,
 by Bob Meyer}

{Additional comments by Gustav Meglicki, ARP, RSSS, ANU, GPO Box 4,
 Canberra, ACT 2601, Australia. Marked by 'G.M.' - NOTE: these initials
 do _not_ stand for 'General Motors'. Occasional references to "Kama
 Sutra" in various comments mean that "Automated Theorem-Proving in
 Non-Classical Logics" by P.B. Thistlewaite, M.A. McRobbie, and R.K. Meyer
 (Pitman, London, 1988) should be consulted.
 G.M.}

label   1;

const
        node_lim    =  4000;
        sublim      = 60;   {internal machine word size; used for
                             aligning records in memory; limit for
                             genrange below}

			    {Formulas are allowed to have up to sublim
			     tokens only. This is apparent after having
			     inspected pre_tabs. Function pre_tabs.getsub,
			     will read line [0..1, 0..200] (see below),
			     and will parse a formula in the line onto
			     a tree. That tree is stored on 
			     pre_tabs.srt [0..sublim] of pre_tabs.tmpsubs,
			     (dots in the names of types and variables
			     indicate nesting!). G.M.}

        axlim       = 300;
        badlim      = 100;
        repmax      = 30;
        dupmax      = 15;
        matsizlim   = 9; {really 10 elements}
        inicost     = 1.20;
	sp	    = ' ';


type    axrange     = 0..axlim;
        badrange    = 0..badlim;
        ternary     = (isin,isout);
        counts      = (scans,made,idemred,curried,swparts,rul2,mat,
                       aclosed ,dclosed,pruned,maxprn,srch,susd,
                       maxwdth,width,maxdpth,depth);
        count_typ   = scans..depth;


                            {SYMBOL DEFINTIONS}

        symtyp      = (fission,conjunction,disjunction,fusion,
                       big_t,sml_t,var1,var2,var3,var4,var5,
                       big_f,sml_f,nvr1,nvr2,nvr3,nvr4,nvr5,nonsymbl);

		      {Up to 5 variables allowed only. These, as it is
		       specified in initprocedure must be 'a', 'b', 'c',
		       'd', and 'e'. The other allowed symbols are
		       '+', 'o', '&', 'V', 'T', 't', 'F', 'f', and '~'.
		       Caret, '^' is also allowed and is internally
		       converted to '~'. This is done in getln. G.M.}

        partrange   = big_t..var5; 
		      {T,t,a,b,c,d,e. G.M.}

        symrange    = fission..nonsymbl; 
		      {+,&,V,o,T,t,a,b,c,d,e, F,f,~a,~b,~c,~d,~e,?. G.M.}

        partset     = packed set of partrange; 
		      {i.e. set of T,t,..,e. G.M.}


                        {MATRIX DEFINITIONS}

        matrange    = 0..matsizlim; {0..9}

        algebra     = record
                        name    : alfa; 
			
				  {alfa is a short (12 char?) string
				   it corresponds to the name of the
				   Matrix used. In fact only one Matrix
				   (CHAIN5) is used in Kripke, and this 
				   is defined in auto_init. G.M.}

                        working : boolean; 
				  {can switch the test on and off. G.M.}

                        negtab  : packed array[0..matsizlim] of matrange;
				 
				  {This is a table which corresponds to
				   negation. G.M.}

                        idfis,    {auxilliary table. G.M.}
                        contab,   {conjunction. G.M.}
                        distab,   {disjunction. G.M.}
                        fustab,   {fusion. G.M.}
                        fistab    {fission. G.M.}
				: packed array[0..matsizlim,0..matsizlim] of 
				  matrange;

                        ldes,     {Lowest designated value. G.M.}
                        siz     : matrange; {Size of matrix. G.M.}
                        des     : packed set of matrange; 
				  {Set of designated values. G.M.}
                      end;


                            {MULTISET DEFINITIONS}

        genrange    = 0..sublim;    {limit of distinct generators; ideally
                                     equal to some multiple of word sizes}
				    {The 'generators' of a multiset are all
				     the members of the set, the multiset is
				     spanned over. sublim = 60. 
				     So, we can have only up to 60 generators
				     in general. G.M.}

        duprange    = 0..dupmax;    {number of copies of a generator; ideally
                                     a power of 2}
				    {dupmax = 15. G.M.}

        reprange    = 1..repmax;    {number of distinct generators in any
                                     given multiset; ideally, duprange x
                                     repmax should occupy a full word}
				    {repmax = 30. So, we can only have
				     up to 30 generators out of 60 available
				     in every multiset.  G.M.}

        repray      = packed array[1..repmax] of duprange;
        genray      = packed array[1..repmax] of genrange;

		      {A multiset can be represented by two arrays: the first
		       of type genray lists the generators. The second,
		       corresponding array lists the number of occurrences
		       of a given generator (same index) in the multiset.
		       There can be only up to 30 different generators
		       per multiset (out of 60 available) and they cannot 
		       repeat more than 15 times. G.M.}

        genset      = packed set of genrange; {A set of generators. G.M.}

        node_state = (failed,open,initial,theorem,linked);

        node_ptr    = 0..node_lim;

		      {The nodes are all packed into array of nodes.
		       The tree structure is then constructed by the means
		       of integer indexes pointing to various nodes.
		       The tree cannot be larger than up to node_lim
		       at a time (which is 4000). This approach may
		       have to be modified in favour of dynamic memory
		       structuring and allocation. On modern systems,
		       this more versatile approach is very efficient. G.M.}

        node        = packed record
                        lmsk, {?. G.M.}
                        rmsk, {?. G.M.}
                        reps    : repray; {arrays which contain information
					   about repetitiveness of various
					   generators from gens. G.M.}

                        gens    : genset; {this is a set of generators. Because
					   the elements of the set are not
					   ordered it's not clear how items
					   in reps correspond to elements of
					   gens. Unless lexicographic ordering
					   takes place somewhere. ?. G.M.}

                        up      : node_ptr;
                        lp,rp   : node_ptr;
                        pc      : genrange; {principal connective?. G.M.}
                        state   : node_state; 

				  {failed, open, initial, theorem, linked.
				  G.M.}

                        stage   : 0..6;   {0=initial,1=no-w,etc.}
					  {?. G.M.}
                        lkey,             {?, all items. G.M.}
                        rkey,
                        locmax,
                        pcloc   : 0..repmax; {repmax = 30. G.M.}
                        failcnt : integer; {?. G.M.}
                      end;



var
	ht,nul,vt   : char;
		      {ht = \t; null = \0; vt = \013 (vertical tab). 
		       These are initialized in initprocedure. 
		       ht is used in memchk, info and error. vt is
		       used in getln only. nul is used in autoinit,
		       getsym, and in print_form. G.M.}

        bad_g       : array[0..badlim] of genset;
		      {badlim = 100; genset = set of generators;
		       used in add_baddy and try_proof.scan.not_baddy. 
		       Failed nodes are being marked and accounted for.
		       The bad formulas are then stored in bad_g and bad_r
		       (see below). G.M.}

        axiom_g     : array[0..axlim] of genset;
		      {axlim = 300; genset = set of generators;
		       used in add_theorem, pre_tabs.maktab.add_init, 
		       pre_tabs.maktab, try_proof.scan.chk_theorem. G.M.}

        bad_r       : array[0..badlim] of repray;
		      {badlim = 100; repray = repetition array;
		       used in add_baddy and try_proof.scan.not_baddy. G.M.}

        axiom_r     : array[0..axlim] of repray;
		      {axlim = 300; repray = repetition array;
		       used in add_theorem, pre_tabs.maktab.add_init,
		       pre_tabs.maktab, and try_proof.scan.chk_theorem. G.M.}

        root        : node_ptr;         {base of pst.}
		      {Base of proof search tree. Used in memchk, pre_tabs,
		       inform, try_proof.scan.curry_fail, try_proof, and
		       main. G.M.}

        tree        : array[0..node_lim] of node;
		      {Proof search tree. In fact this is NOT the pst as such.
		       This is the space in which the tree itself can grow.
		       Used in:
			autoinit
			assign
			purge
			memchk
			add_baddy
			add_theorem
			pre_tabs.maktab.add_init
			print_mset
			inform
			try_proof
				matrix_test
				scan
					part_chk
					curry_fail
					not_baddy
					chk_theorem
					getpc
					defson
					create
					apply_fusion
						getright
							eql
						getleft
							getl
						check_fusions
				dpth_stat
			main
		       G.M.}

        free_node   : node_ptr; {integers pointing to appropriate entry in 
				 tree. G.M.}
        node_cnt    : node_ptr;
        tmp_node    : node_ptr;

        baddy_cnt   : badrange; 
		      {0..badlim; badlim = 100. 
		       Used in add_baddy, pre_tabs.maktab, info, and
		       try_proof.scan.not_baddy. G.M.}
        axiom_cnt,
        theorem_cnt : axrange; {0..axlim; axlim = 300. G.M.}
        cc          : count_typ; {scans..depth. G.M.}
        count       : array[scans..depth] of integer;
		      {scans..depth really go like
		       scans, made, idemred, curried, swparts, rul2, mat,
		       aclosed, dclosed, pruned, maxprn, srch, susd,
		       maxwdth, width, maxdpth, depth.
		       So, things such as cc go through all these operations
		       in exactly the order shown above. G.M.}

        matrix      : algebra;
		      {used in autoinit (where it is initialized to
		       CHAIN5), try_proof.matrix_test.eval,
		       try_proof.matrix_test.f_eval, and try_proof.matrix_test.
		       As it has been pointed above, only one matrix is used
		       by kripke. G.M.}
        chkpt,
        endtime,
        strtime     : integer;


        token       : packed array[1..2] of char;
		      {Tokens are things such as 'o ', or 'a '. 2 characters
		       are needed for negated variables, e.g., '~a'. 
		       '^a' is converted to '~a' by getln. 
		       Used only by pre_tabs.getsym and should be made
		       an internal variable of that function. G.M.}

        tok         : array[fission..nonsymbl] of packed array[1..2] of char;
		      {This array makes the correspondence between ascii
		       tokens and their meanings. Initialized in initprocedure
		       to:
				fission		-> '+ '
				fusion		-> 'o '
				conjunction	-> '& '
				disjunction	-> 'V '
				big_t		-> 'T '
				sml_t		-> 't '
				big_f		-> 'F '
				sml_f		-> 'f '
				var1		-> 'a '
				var2		-> 'b '
				var3		-> 'c '
				var4		-> 'd '
				var5		-> 'e '
				nvr1		-> '~a'
				nvr2		-> '~b'
				nvr3		-> '~c'
				nvr4		-> '~d'
				nvr5		-> '~e'
		      tok [nonsymbl] is never used.
		      One wonders if it wouldn't be more suitable to do
		      manipulations on the tokens directly, and equip the
		      program in knowledge about special tokens which
		      correspond to operations. 
		      Used in initprocedure, auto_init, pre_tabs.getsym,
		      and print_form. G.M.}

        dual        : array[fission..nonsymbl] of symtyp;
		      {symtyp = fission..nonsymbl; every one of these
		       has its dual. dual is initialized by initprocedure
		       to yield:
				+	<->	o
				&	<->	V
				T	<->	F
				t	<->	f
				a	<->	~a
				b	<->	~b
				c	<->	~c
				d	<->	~d
				e	<->	~e
				?	<->	?	(nonsymbl)
		      Used in initprocedure, auto_init,
		      pre_tabs.maktab.contradict, try_proof.matrix_test.eval. 
		      G.M.}

        pp,np       : packed array[fission..nonsymbl] of partset;
		      {partset = set of [T,t,F,f,a,b,c,d,e]. pp and np are
		       arrays of constants and variables. 
		       Both are used in auto_init, and pre_tabs.maktab.fill_tab
		       only. G.M.}

        asc         : char;
		      {Used in getln only. Should be a local variable. G.M.}

        subs        : array[0..sublim] of genset;            {subs of sub}
		      {sublim = 60, genset = set of [0..60] (probably a
		       set of generators). 
		       Used only in pre_tabs.maktab.fill_tab, and 
		       try_proof.scan.part_chk. G.M.}

        posp,negp   : array[0..sublim] of partset;           {pos/neg parts}
		      {sublim = 60, partset = set of [T,t,F,f,a,b,c,d,e].
		       Used in pre_tabs.maktab.fill_tab, try_proof.matrix_test,
		       and try_proof.scan.parth_chk. G.M.}

        sym         : array[0..sublim] of symtyp;            {symbol}
		      {sublim = 60, symtyp = fission (+) .. nonsymbl.
		      Used in initprocedure, pre_tabs.maktab.fill_tab,
		      pre_tabs.maktab.contradict, pre_tabs.maktab, print_form,
		      print_mset, try_proof.matrix_test.eval,
		      try_proof.scan.getpc, and try_proof.scan. G.M.}

        contra,                                             {contradict. sub}
		      {Used in pre_tabs.maktab only. G.M.}

        lsub,rsub   : array[0..sublim] of genrange;          {left/right sub}
		      {These two variables are used to store left and right
		       parts of a formula. The easiest way to see how it is
		       done is to inspect procedure print_form, and read
		       the G.M. comments there.
		       lsub is used in pre_tabs.maktab.fill_tab,
		       pre_tabs.maktab.contradict, pre_tabs.maktab, print_form,
		       try_proof.matrix_test.eval,
		       try_proof.scan.apply_fusion.getleft.getl, and in
		       try_proof.scan.
		       rsub is used in the same routines with the exception of
		       ...getleft.getl. Instead it is used in
		       try_proof.scan.apply_fusion.getright. Note the lack of
		       symmetry between right and left. G.M.}

        subcost     : array[0..sublim] of real;              {cost}
		      {subcost is used only in one place: in pre_tabs.maktab.
		       A value is put into it, and then it is never used.
		       Looks like a spurious variable. G.M.}

        bigt,smlt,    {bigt contains a location of 'T' in sym, likewise for
		       smlt, bigf, and smlf. 
		       bigt is used in pre_tabs.maktab and in
		       try_proof.scan.chk_theorem. smlt is only used in
		       pre_tabs.maktab. Some values are put into it, but
		       then they are never taken out of it. Looks like a
		       spurious variable. bigf is used in pre_tabs.maktab
		       only. smlf is used in pre_tabs.maktab and then in
		       try_proof.scan.defson.
		       G.M.}
        bigf,smlf,
        rootsub,      {rootsub is used in memchk, pre_tabs.maktab, pre_tabs,
		       and in try_proof. G.M.}

        submax      : genrange;
		      {submax is used in pre_tabs.getsub, pre_tabs.maktab,
		       pre_tabs, try_proof.scan.defson, try_proof.scan.create,
		       try_proof.scan.apply_fusion.getright, and
		       ... getleft.getl. It is used most often as an upper
		       limit of various loops. G.M.}

        r2parts,
        atparts,
        strongparts,  {r2parts, atparts, and strongparts are used only in
		       pre_tabs.maktab and in try_proof.scan.part_chk. G.M.}

        inverts     : genset; {set of generators ([0..60]). G.M.}
		      {inverts is used only in pre_tabs.maktab. G.M.}

        linlen      : array[0..1] of integer;
		      {The input line is split into two parts: the first
		       one contains a formula, the second one contains
		       a comment. linlen contains information about the
		       lengths of both parts. 
		       Used in getln, pre_tabs, and in info. G.M.}

        linloc,       {Probably points to the location in the input string.
		       Used in getln, pre_tabs.getsym, and pre_tabs. G.M.}

        lino,         {Line number. Used in getln and main. G.M.}

        x,y         : integer;
		      {Loop counters. Auxilliary integers. G.M.}
			
        line        : packed array[0..1,0..200] of char;
		      {Input line. Can be split in two parts by ';'.
		       The stuff after the semicolon is a comment. 
		       line [1... is used only in info. 
		       line as such is used in getln, pre_tabs.getsym, and
		       in info. G.M.}

        remat,matnam,
        nam         : alfa; 
		      {remat is not used anywhere at all. It is a spurious
		       variable.  matnam is only used in auto_init and it 
		       holds the name of the matrix, i.e., CHAIN5, but it 
		       is not recalled anywhere. nam is used in inform to get
		       the time via time(nam). G.M.}

        endform     : boolean;
		      {endform is set to true if '*' is registered on the
		       input line. It is used only in getln and main. G.M.}


procedure initprocedure;

{Initialize ht, v, nul, sym[0], duals, and tok. 
 Called only once from main.
 G.M.}

    begin

    ht:=chr(9); vt:=chr(11); nul:=chr(0);

    sym[0]:=nonsymbl;

    dual[fission]:=fusion;          dual[var1]:=nvr1;   dual[nvr1]:=var1;
    dual[fusion]:=fission;          dual[var2]:=nvr2;   dual[nvr2]:=var2;
    dual[conjunction]:=disjunction; dual[var3]:=nvr3;   dual[nvr3]:=var3;
    dual[disjunction]:=conjunction; dual[var4]:=nvr4;   dual[nvr4]:=var4;
    dual[nonsymbl]:=nonsymbl;       dual[var5]:=nvr5;   dual[nvr5]:=var5;
    dual[big_t]:=big_f;             dual[big_f]:=big_t;
    dual[sml_t]:=sml_f;             dual[sml_f]:=sml_t;

    tok[fission]:=     '+ ';      tok[var1]:='a '; tok[nvr1]:='~a';
    tok[fusion]:=      'o ';      tok[var2]:='b '; tok[nvr2]:='~b';
    tok[conjunction]:= '& ';      tok[var3]:='c '; tok[nvr3]:='~c';
    tok[disjunction]:= 'V ';      tok[var4]:='d '; tok[nvr4]:='~d';
    tok[big_t]:=       'T ';      tok[var5]:='e '; tok[nvr5]:='~e';
    tok[sml_t]:=       't ';
    tok[big_f]:=       'F ';
    tok[sml_f]:=       'f ';

    end; {initprocedure}

function max(x,y : integer) : integer;

{Return the larger of two integers. 
 Called from autoinit, info, try_proof.dpthstat, and try_proof.
 G.M.}

	begin
	if x>y then max:=x else max:=y;
	end; {max}

function min(x,y : integer) : integer;

{Return the smaller of two integers. 
 Called from autoinit and try_proof.matrix_test.f_eval.
 G.M.}

	begin
	if x<y then min:=x else min:=y;
	end; {min}



procedure error(n,loc:integer); forward;

procedure auto_init(block:integer);

{if block = 0: Initialize pp[s] to [s] if s is positive, and to empty set
	       otherwise. np[dual[s]] initialize to [s], if s is negative, 
	       and to empty set otherwise.
	       Initialize matrix to hold CHAIN5.
 if block = 1: Link the nodes into a ring, leaving the left pointers
	       dangling. Mark node 1 as the first free node (free_node = 1), 
	       mark number of available nodes (node_cnt) as node_lim (4000).
 Called from main (both for block = 0 and block = 1) and from memchk (with
 block = 1 only, i.e., it relinks the nodes into the ring and marks tree as
 empty.
 G.M.}

    var     s   : symtyp;
    begin
    case block of
     0: begin       {initialise symbol table}
        for s:=fission to nonsymbl  do begin np[s]:=[]; pp[s]:=[] end;
        for s:=big_t to var5 do
          begin pp[s]:=[s]; np[s]:=[]; pp[dual[s]]:=[]; np[dual[s]]:=[s] end;
        for s:= fission to sml_f do tok[s,2]:=nul;
        matnam:='CHAIN5    ';
	with matrix do
		begin
		name	:= 'CHAIN5    ';
		working	:= true;
		siz	:= 4;
		des	:= [1,2,3,4];
		ldes	:= 1;
		negtab[4]:=0; negtab[3]:=1; negtab[2]:=2; negtab[1]:=3;
		negtab[0]:=4;
		for x:=0 to siz do for y:=0 to siz do
			begin
			contab[x,y]:=min(x,y);
			distab[x,y]:=max(x,y);
			fustab[x,y]:=4;
			end;
		for x:=0 to siz do
			begin
			fustab[0,x]:=0;
			fustab[x,0]:=0;
			fustab[x,1]:=x;
			fustab[1,x]:=x;
			idfis[x,1]:=x;
			end;
		fustab[2,2]:=3;
		for x:=0 to siz do for y:=0 to siz do
			fistab[x,y]:=negtab[fustab[negtab[x],negtab[y]]];
		for x:=0 to siz do for y:=2 to siz do
			idfis[x,y]:=fistab[x,idfis[x,y-1]];
		end;
        end;
     1: begin       {initialise tree of node ptrs}
        for x:=1 to (node_lim-1) do tree[x].rp:=x+1;
        tree[node_lim].rp:=0; free_node:=1; node_cnt:=node_lim;
        end;
    end; {case}
    end; {autoinit}




{----- UTILITY ROUTINES ------------------------------------------------------}


function xp(n:integer) : char;

{Write integer n left justified and return space. Thus: write (xp(n)), for
 instance, will produce a left justified n followed by one space.
 Called from add_baddy, inform, info,
 try_proof.scan.apply_fusion.check_fusions, and try_proof.
 G.M.}

    var x,y : integer;
    begin
    x:=0; y:=n;
    repeat x:=x+1; y:=y div 10; until y=0;
    write(n:x);
    xp:=' ';
    end; {xp}

procedure getln;

{Read input line: echo what is read to stdout adding the line number,
 convert '^', to '~', set up endform to true if '*' is encountered, ignore 
 spaces. If ';' is encountered switch to filling up line [1,?]. line [1,?] 
 will not be processed, but it will be printed on stdout by info.
 Called only once from pre_tabs.
 G.M.}

    var     comm : integer;
        finished : boolean;
    begin
    linloc:=0; linlen[0]:=0; linlen[1]:=0; comm:=0; lino:=lino+1;
    finished:=false;
    repeat
                writeln(vt); for x:=1 to 125 do write('%'); writeln;
                finished:=true; write('Disk input',lino:4,': ');
                repeat  read(asc);
			write(asc);
                        if asc='^' then asc:='~';
                        case asc of
                        '*'      : begin endform:=true; finished:=true end;
                        sp       : begin end;
                        ';'      : comm:=1;
                        otherwise: begin
                                   linlen[comm]:=linlen[comm]+1;
                                   line[comm,linlen[comm]]:=asc;
                                   end;
                        end;  {case}
                until eoln;
        until finished;
    writeln; flush(output);
    end; {getln}

procedure trans(gs:genset;var gr:genray); 

{For every generator found in gs, put its number in the next free slot of gr.
 Pad the rest of gr with 0s. In other words, translate a set onto an array.
 The elements of the set will be ordered in the array according to their
 numbers (genset is really a set of a subrange of integers).
 Called from print_mset, try_proof.matrix_test, and try_proof.scan.
 G.M.}

	var	x,y : integer;
	begin
	y:=0;
	for x:=0 to sublim do if x in gs then
		begin y:=y+1; gr[y]:=x end;
	for x:=y+1 to repmax do gr[x]:=0;
	end; {trans}

function frstpc(gs:genset;var p:genrange) : boolean; 

{Find the lowest generator in gs that is larger than p. Return 'true'
 if such a generator exists and put it in p. Return 'false' if there is
 no generator in gs larger than p. That 'pc' stands for "Principal Constituent"
 in case you wondered about it. See Definition 2.2, page 28 of Kama Sutra for
 more information.
 Called only from try_proof.scan.getpc.
 G.M.}

	label	36;
	var	x : integer;
	begin
	frstpc:=false;
	for x:=(p+1) to sublim do if x in gs then
		begin frstpc:=true; p:=x; goto 36 end;
   36:  end; {frstpc}

procedure assign(var p:node_ptr);

{This is our own local version of standard Pascal "new" (or "malloc" in C).
 The first available node pointer is put in p. Because the nodes were initially
 linked to form a ring, the next free node is put into free_node and then
 the node pointed to by p is reinitialized. An unused copy of a virgin node
 is always kept on tree[0]. If there are no free nodes error is flagged.
 '1' means: "node space exhausted"; '2' means: "trying to assign NULL node".
 Error flag actually does not abort the program, but it does abort processing
 of the current formula.
 Called from:
	try_proof
		scan.create
		scan.apply_fusion
				getleft.getl
 G.M.}

    var x:integer;
    begin
    x:=p;
    p:=free_node; free_node:=tree[free_node].rp;  tree[p]:=tree[0];
    if free_node=0 then error(1,x); node_cnt:=node_cnt-1;
    if p=0 then error(2,x);
    end; {assign}

procedure purge(var p:node_ptr);

{Free the whole branch that has its root at p. As the nodes are freed they
 are being linked from the left into the list of free nodes. 
 Called from:
	pre_tabs
	try_proof.scan
			getpc
			apply_fusion
				getleft
				check_fusions
 G.M.}

    var     x : node_ptr;
    begin
    if p<>0
       then with tree[p] do
            begin
            x:=lp; purge(x); x:=rp; purge(x);
            node_cnt:=node_cnt+1;
            rp:=free_node;
            free_node:=p;
            p:=0
            end;
    end; {purge}

procedure memchk;

{Check if information about the number of available nodes kept on
 node_cnt agrees with the number of nodes in the linked list beginning
 with free_node and ending with 0. If not reinitialize all nodes in
 the "tree" array to form the ring (done by auto_init(1)).
 Called only from error.
 G.M.}

    begin
    write(ht,'Memory check: '); x:=0; y:=free_node;
    while y<>0 do begin x:=x+1; y:=tree[y].rp end;
    if x=node_cnt
       then writeln('consistent')
       else begin
            writeln('CORRUPTED - recovering'); auto_init(1);
            root:=0; rootsub:=0;
            end;
    end; {memchk}


procedure print_mset(n:node_ptr;prtsb:boolean;dpth:integer); forward;

        procedure add_baddy(l:integer);

	{This procedure maintains the list of baddies - whatever these
	 might be. A node in the tree, marked by someone else as bad,
	 is passed to it. This node is then checked agains the list of
	 bad multisets which is maintained in bad_g[0..100] (the generators)
	 and in bad_r[0..100] (the repetitions). If the sets match(i.e. 
	 tree[l].gens = bad_g[n]) then the numbers in bad_r and tree[l].reps 
	 are compared.  If for any generator in a matched set the repetition 
	 number in tree[l].reps is less than its corresponding repetition 
	 number in bad_r, then isin flag is raised. If the opposite happens, 
	 the isout flag is raised. There are now four possibilities: 
	 cont = [], cont = [isin], cont = [isout], and cont = [isin, isout]. 
	 If cont = [isout] then the procedure is aborted without any change 
	 to bad_g and bad_r.  If there are two multisets A and B which have 
	 the same set of generators (i.e. they are cognate), and for every 
	 generator a: card (a) in A <= card (a) in B then we say that A is 
	 strongly contained in B. This means that if a multiset is passed to
	 add_baddy such that there is already a multiset present in the
	 baddies register which is strongly contained in the new multiset,
	 that new multiset is not considered new and is thrown away.
	 NOTE one subtlety here: the strong containment has <= sign, i.e.
	 B / B (/ is the notation for strong containment of multisets),
	 whereas in this procedure a sharp inequality (<) is used!
	 This may lead to identical copies of the bad multisets being
	 added to the list, unless appropriate check is made elsewhere. 
	 If cont = [isin], i.e. the new multiset is strongly contained
	 in one of the multisets in the baddies register then the old baddy
	 is replaced with the new one. In other cases, i.e. [isin, isout]
	 or [] (this will be possible only if A = B, or if no match
	 for set A exists in the baddies register) the new multiset will be
	 appended to the list of baddies. Consult Kama Sutra, Definition 2.1,
	 page 27 for more information.
	 ALSO NOTE that checking for array bounds overflow (i.e. baddy_cnt <
	 badlim) occurs only in one place and no error is marked if the 
	 overflow occurs! In that case, the new baddy is NOT added to the
	 list but nothing else happens. This may lead to potential errors
	 in case of long proofs.
	 Called from tryp_proof.scan only.
	 G.M.}



            label 36;
            var n:genrange;
                x:genset;
                y:repray;
                z:reprange;
                cont: packed set of ternary;
                freeloc:integer;
            begin
            x:=tree[l].gens; y:=tree[l].reps;
            freeloc:=0;
            for n:=1 to baddy_cnt do
                begin
                cont:=[];
                if bad_g[n]=x
		   {card(x) is a Cardinality of x. G.M.}
                   then for z:=1 to card(x) do
                            if y[z]<bad_r[n][z] then cont:=cont+[isin]
                     else   if y[z]>bad_r[n][z] then cont:=cont+[isout]
                   else if (bad_g[n]=[]) and (freeloc=0) then freeloc:=n;
                if cont=[isout]
                   then goto 36
                   else if cont=[isin]
                           then begin
                                if freeloc=0 then freeloc:=n;
                                bad_g[n]:=[]; bad_r[n]:=tree[0].reps;
                                end;
                end;
            if freeloc=0
               then begin
                    if baddy_cnt<badlim
                       then begin baddy_cnt:=baddy_cnt+1;
                                  write('BADDY ',xp(baddy_cnt),'added : ');
                                  print_mset(l,false,0);
                                  bad_g[baddy_cnt]:=x;
                                  bad_r[baddy_cnt]:=y;
                            end
                    end
               else begin
                    bad_g[freeloc]:=x; bad_r[freeloc]:=y;
                    write('BADDY ',xp(freeloc),'replaced : ');
                    print_mset(l,false,0);
                    end;
        36: end; {add_baddy}


        procedure add_theorem(l:integer);

	{This procedure works in a way which is almost identical
	 to add_baddy. The differences are as follows: the multisets
	 passed to add_theorem are compared against the multisets stored
	 in axiom_g/axiom_r pair. The theorem is not considered worth
	 adding or replacing the existing one in the list if it _is_
	 strongly contained in the one from the list. It will replace
	 the one in the list if it strongly contains it. So, this is
	 the opposite to the baddy strategy. Again the equal case is
	 not checked and if theorem_cnt >= axlim (= 300) then nothing
	 happens apart from the theorem not being appended to the list.
	 I.e., kripke will quietly go on, as if nothing happened. This 
	 may lead to some errors in case of long proofs.
	 Called from try_proof.scan only.
	 G.M.}

            label 35;
            var n:genrange;
                x:genset;
                y:repray;
                z:reprange;
                cont: packed set of ternary;
                freeloc:integer;
            begin
            x:=tree[l].gens; y:=tree[l].reps;
            freeloc:=0;
            for n:=1 to theorem_cnt do
                begin
                cont:=[];
                if axiom_g[n]=x
                   then for z:=1 to card(x) do
                            if y[z]<axiom_r[n][z] then cont:=cont+[isin]
                     else   if y[z]>axiom_r[n][z] then cont:=cont+[isout]
                   else if (axiom_g[n]=[]) and (freeloc=0) then freeloc:=n;
                if cont=[isin]
                   then goto 35
                   else if cont=[isout]
                           then begin
                                if freeloc=0 then freeloc:=n;
                                axiom_g[n]:=[]; axiom_r[n]:=tree[0].reps;
                                end;
                end;
            if freeloc=0
               then begin
                    if theorem_cnt<axlim
                       then begin theorem_cnt:=theorem_cnt+1;
                                  axiom_g[theorem_cnt]:=x;
                                  axiom_r[theorem_cnt]:=y;
                            end
                    end
               else begin axiom_g[freeloc]:=x; axiom_r[freeloc]:=y end;
        35: end; {add_theorem}

{----- INPUT ROUTINES FOR FORMULAS  and PRE-ANALYSIS -------------------------}


function pre_tabs : boolean;

{This function is commented just in front of its main begin-end block. G.M.}

    type    tmpsubs = record s:symtyp; loc,l,r:genrange; c:real end;
    var     x       : integer;
            srt     : array[0..sublim] of tmpsubs;
            realoc  : packed array[0..sublim] of genrange;

    function getsym(var sy:symtyp) : boolean; {pre_tabs.getsym}

    {This function will fetch the next character from the line and it will
     return a corresponding scalar in sy (i.e. fission..nonsymbl).
     getsym returns true if the character could be matched against
     any of the registered tokens and false otherwise.
     Called from pre_tabs.getsub only.
     G.M.}

        begin
        linloc:=linloc+1;
        token[1]:=line[0,linloc];
        if token[1]='~'
           then begin linloc:=linloc+1; token[2]:=line[0,linloc] end
           else token[2]:=nul;
        sy:=fission;
	while not((sy=nonsymbl) or (tok[sy]=token)) do sy:=succ(sy);
        getsym:=sy<>nonsymbl;
        end; {pre_tabs.getsym}

    function getsub(var n:genrange) : boolean; {pre_tabs.getsub}

    {See comments just in front of the main begin for this function
     G.M.}

        var     tmp : tmpsubs;
		      {tmpsubs is a local type defined within the body
		       of pre_tabs only. It is a
			  record
			     s : symtyp; (i.e. fission..nonsymbl)
			     loc, l, r : genrange; (i.e. 0..60)
			     c : real; (cost)
			  end;
		       G.M.}

        function insert : genrange;

	{This function will append tmp to the list of tmpsubs held on
	 srt, if tmp has not been found in srt yet. In this case insert
	 will return the index of tmp in srt, and it will set submax
	 to that index. If tmp has been found in srt, then insert will
	 be still set to the index of tmp in srt, but submax will not
	 be modified. It follows from the above that submax tells
	 how many slots in srt have been filled.
	 Called from pre_tabs.getsub only.
	 G.M.}

            begin
            x:=0;
            repeat
                 x:=x+1; tmp.loc:=x;
                 if x>submax
                    then begin srt[x]:=tmp; submax:=x; insert:=x; x:=-1 end
                    else if srt[x]=tmp then begin insert:=x; x:=-1 end;
            until x=-1;
            end; {pre_tabs.getsub.insert}

    {This function will fetch the next item from line. getsym will convert
     that item to appropriate symbol (if possible), and then getsub will
     check if it is an operator. If not, then the corresponding record
     (tmp) will be inserted into srt (with insert). The cost of the record 
     will be set to inicost and left and right pointers will be set to NIL. 
     If the symbol IS an operator, then getsub will call itself recursively
     for left and right sides of the operator (in this order). The cost
     for the symbol will then be calculated according to the type of the
     symbol and the costs of its left and right branches. Finally the
     corresponding record (tmp) will be inserted into srt (with insert).
     getsub will return true if the formula is correct and false otherwise.
     (It must have all its right and left parts in place to be correct)
     Parameter n (which is var) will be set to the index of tmp in srt.
     NOTE that for this function to work the input formula MUST be given
     in the stack notation with operators first, i.e., reversed reverse
     Polish. At the end of this function every token from the formula
     will have its index in srt. But every literal will appear in 
     srt ONLY ONCE!!! Remember that insert throws away tmp if there is
     already a full match found in srt. 
     Called from pre_tabs only.
     G.M.}

        begin {pre_tabs.getsub}
        getsub:=false;
        with tmp do 

	{Put the next symbol from the line into tmp.s (getsym will
	 translate ascii into the corresponding symbol). If the input
	 was correct then proceed.
	 G.M.}

        if getsym(s) then
           if s in [fission..fusion] {I.e., if s is an operator. G.M.}
              then begin
                   if getsub(l) 

		      {This will really process next character from
		       line, but it will be put into tmp.l, because in
		       traditional notation it corresponds to the left
		       hand side of the fission..fusion operators. G.M.}

                      then if getsub(r)
			      
			      {This, in turn, will fetch the right hand
			       side of the operator. G.M.}

                              then  begin
                                    getsub:=true; 

				 	{The subformula is correct.  G.M.}

                                    case s of
                                    fission     : if srt[l].c > srt[r].c
						     then c:=srt[l].c
						     else c:=srt[r].c;

					{Make the cost associated with the
					 present symbol (fission) equal to 
					 the larger of the costs of its
					 two sides. G.M.}

                                    fusion      : c:=srt[l].c*srt[r].c;

					{Make the cost associated with the
					 present symbol (fusion) equal
					 to the product of the costs of its
					 two sides. G.M.}

                                    conjunction : if srt[l].c < srt[r].c
						     then c:=srt[l].c
						     else c:=srt[r].c;

					{Make the cost associated with the
					 present symbol (conjunction) equal
					 to the lesser of the costs of its
					 two sides. G.M.}

                                    disjunction : c:=srt[l].c+srt[r].c;

					{Make the cost associated with the
					 present symbol (disjunction) equal
					 to the sum of the costs of its two
					 sides. G.M.}

                                    end;
                                    n:=insert;
					
					{Insert this record (tmp) into
					 the srt array. Set n to the returned 
					 index of tmp in srt.
					 G.M.}

                                    end;
                   end
              else 

		 {I.e., if s is NOT an operator. G.M.}

	      begin l:=0; r:=0; getsub:=true; c:=inicost; n:=insert end

                 {If s is not an operator then it will not have left and
		  right sides.}

        end; {pre_tabs.getsub}


    procedure maktab; {pre_tabs.maktab}
        var     x,y,z   : genrange;

        procedure switch(n:genrange); {pre_tabs.maktab.switch}

	{Transpose n and n+1 elements of srt.
	 Called from pre_tabs.maktab.bubble and pre_tabs.maktab.
	 G.M.}

            var y:tmpsubs;
            begin
            y:=srt[n]; srt[n]:=srt[n+1]; srt[n+1]:=y;
            end; {pre_tabs.maktab.switch}

        procedure bubble(n:genrange; key:integer); {pre_tabs.maktab.bubble}

	{Bubble sort the first n entries of srt with respect to cost of 
	 entries for key = 1 and with respect to ordering of symbols for 
	 key = 2.
	 Called from pre_tabs.maktab only.
	 G.M.}

            begin
            while n>1 do
                begin
                n:=n-1;
                case key of
                 1: if srt[n].c > srt[n+1].c
                       then switch(n) else n:=0;
                 2: if srt[n].s > srt[n+1].s
                       then  switch(n) else n:=0;
                end;
                end;
            end; {pre_tabs.maktab.bubble}

        procedure fill_tab(n:genrange); {pre_tabs.maktab.fill_tab}

	{This function will fill positive-part, negative-part and subformula
	 tables. The starting point is n (the generator), pp[sym[n]], 
	 np[sym[n]], lsub[n], and rsub[n]. The procedure will recursively
	 descend down lsub and rsub (in this order) until the corresponding
	 lsub[lsub[lsub... is 0. All of these are filled within the main
	 maktab itself.
	 Called from pre_tabs.maktab only.
	 G.M.}

            begin
            if n<>0
               then begin
                    fill_tab(lsub[n]); fill_tab(rsub[n]);

		    {The special clauses checking if the symbol is sml_t
		     or sml_f have been suggested by the illustrious P.T. 
		     after a heated debate on why KRIPKE would refute
		     (t o t). G.M.}

		    if sym[n] = sml_t then begin
                       posp[n]:=posp[lsub[n]]+posp[rsub[n]]+pp[sym[n]];
                       negp[n]:=negp[lsub[n]]+negp[rsub[n]]+pp[sym[n]];
                       subs[n]:=subs[lsub[n]]+subs[rsub[n]]+[n];
		    end
		    else if sym[n] = sml_f then begin
                       posp[n]:=posp[lsub[n]]+posp[rsub[n]]+np[sym[n]];
                       negp[n]:=negp[lsub[n]]+negp[rsub[n]]+np[sym[n]];
                       subs[n]:=subs[lsub[n]]+subs[rsub[n]]+[n];
		    end
		    else begin
                       posp[n]:=posp[lsub[n]]+posp[rsub[n]]+pp[sym[n]];
                       negp[n]:=negp[lsub[n]]+negp[rsub[n]]+np[sym[n]];
                       subs[n]:=subs[lsub[n]]+subs[rsub[n]]+[n];
                    end
	       end
            end; {pre_tabs.maktab.fill_tab}

        function contradict(n,m:genrange) : boolean;
	{pre_tabs.maktab.contradict}

	{Recursively check if the two formulas represented by their
	 principal generators n and m contradict one another. Do that
	 in the following way: first check if m is the dual of n. If
	 so then run contradict on left and right sides of n and m.
	 Continue until you hit sym[lsub[lsub..[lsub[n]]..]] = nonsymbl.
	 Called from pre_tabs.maktab only.
	 G.M.}

            begin
            if sym[n]=dual[sym[m]]
               then if sym[n]=nonsymbl
                       then contradict:=true
                       else contradict:=contradict(lsub[n],lsub[m]) and
                                        contradict(rsub[n],rsub[m])
               else contradict:=false;
            end; {pre_tabs.maktab.contradict}


        procedure add_init(x,y:genrange); {pre_tabs.maktab.add_init}

	{Find next empty slot in the list of axioms, initialize it,
	 add x and y to the set axiom_g[theorem_cnt] and set up
	 the repetition counters in the corresponding axiom_r[theorem_cnt].
	 NOTE that this procedure does not check for array bounds.
	 Up to 300 axioms can be placed in axiom_[rg].
	 Called from pre_tabs.maktab only.
	 G.M.}

            var n:genrange;
            begin
            theorem_cnt:=theorem_cnt+1;
            axiom_r[theorem_cnt]:=tree[0].reps;
            axiom_r[theorem_cnt][1]:=1;
            axiom_g[theorem_cnt]:=[x];
            if y<>0 then begin
                         axiom_g[theorem_cnt]:=axiom_g[theorem_cnt]+[y];
                         axiom_r[theorem_cnt][2]:=1;
                         end;
            end; {pre_tabs.maktab.add_init}

	{This largish procedure will sort entries in srt according to
	 cost and then according to symbol ordering. The .l and .r pointers
	 will be reallocated. Information contained in srt is then transferred
	 to global tables, most notably to sym (symbols corresponding to
	 generators), subcost (cost of the generator), lsub (pointers to
	 left parts of formulae, if any), rsub (pointers to right parts of
	 formulae, if any). Once this is done all specific occurrences of
	 basic theorems for the variables used in the input formula will
	 be derived and stored on axiom_g and axiom_r. If <a + ~a> pair is
	 encountered then the record of inverts will be appended to. If
	 disjunction is encountered then the record of strongparts is appended
	 to. If fusion is encountered the record of r2parts will be appended 
	 to. For other symbols the record of atparts is appended to.
	 Called from pre_tabs only.
	 G.M.}

        begin {pre_tabs.maktab}
                        {Sort table entries}
        for z:=1 to 2 do
            for x:=1 to (submax-1) do
                case z of
                    1: if srt[x].c > srt[x+1].c
                          then begin switch(x); bubble(x,z) end;

		       {Sort entries in srt with respect to cost. G.M.}

                    2: if srt[x].s > srt[x+1].s
                          then begin switch(x); bubble(x,z) end;

		       {Sort entries in srt with respect to symbol. G.M.}

                end;
		       {But, guess what happens to .l and .r indexes! They
			now point to completely wrong entries! Alas, the
			original position of records in srt is still
			preserved in .loc, and we'll quickly deal with this
			now. G.M.}

                        {Transfer from tmp-tables to global tables}
	realoc[0]:=0;
        for x:=1 to submax do 
		begin
		y:=srt[x].loc;
		realoc[y]:=x;
		end;

	{So, now realoc[y] contains information about the original
	 position of the records. Consequently, .l and .r will have
	 to be translated by realoc whenever we are to make any use of
	 them. G.M.}

	{Now we transfer information from srt to our global tables:
	 sym (contains symbols corresponding to x), subcost (contains
	 costings corresponding to x), lsub (contains REAL, this time,
	 pointers to the left side of an operator, if x is one), rsub (contains 
	 pointers to the right side of an operator, if x is one). rootsub
	 will be initialized by the preceding call to getsub and will point
	 to the principal connective of the formula. G.M.}

        for x:=1 to submax do
            begin
            sym[x]:=srt[x].s; subcost[x]:=srt[x].c;
            lsub[x]:=realoc[srt[x].l]; rsub[x]:=realoc[srt[x].r];
            end;
        rootsub:=realoc[rootsub];
        fill_tab(rootsub);

                        {Get basic <a,~a> axioms}
			{In other words check for the presence of any a, ~a 
			 pairs in the formula. All possible pairs of generators
			 in the formula are scanned and every <positive,
			 negative> pair found is registered as an axiom. Also,
			 every occurrence of small t is registered as an axiom.
			 In this way KRIPKE generalizes <a,~a> to all other
			 possible pairs of symbols.
			 G.M.}

        axiom_cnt:=0; theorem_cnt:=0; baddy_cnt:=0;
        for x:=1 to submax do contra[x]:=0;
        for x:=1 to submax do
            begin
            if sym[x]=sml_t
               then begin axiom_cnt:=axiom_cnt+1; add_init(x,0);      end;
            for y:=(x+1) to submax do
                if contradict(x,y)
                   then begin
                        contra[x]:=y; contra[y]:=x;
                        axiom_cnt:=axiom_cnt+1;
                        add_init(x,y);
                        end;
            end;

                        {Derive elementary theorems}
			{KRIPKE wouldn't know on its own that a given
			 theorem, expressed in terms of a and b would
			 generalize to all other pairs of variables, e.g.,
			 d and e, or a and c. So we have to go through
			 all the variables used in the input formula and
			 reconstruct the basic theorems for every combination
			 of variables that might arise during processing
			 of the formula. G.M.}

        inverts:=[]; strongparts:=[]; r2parts:=[]; atparts:=[];
        bigt:=0; smlt:=0; smlf:=0; bigf:=0;
        for x:=1 to submax do
            case sym[x] of
            big_f       : bigf:=x;
            sml_f       : smlf:=x;
            big_t       : bigt:=x;
            sml_t       : smlt:=x;
            fission     : begin
                          if contra[lsub[x]]=rsub[x]
                             then add_init(x,0);
                          inverts:=inverts+[x];
                          end;

			  {a + ~a is a theorem. Note, that only the
			   fusion is put in axiom_g/axiom_r. This fusion
			   points to appropriate left and right sides through
			   its corresponding lsub and rsub entries. G.M.}

            disjunction : begin
                          strongparts:=strongparts+[x];
                          if contra[lsub[x]]=rsub[x]
                             then add_init(x,0);

			  {a V ~a is a theorem. G.M.}

                          if contra[lsub[x]]<>0
                                then add_init(x,contra[lsub[x]]);
                          if contra[rsub[x]]<>0
                                then add_init(x,contra[rsub[x]]);

			  {[a V b, ~a] and [a V b, ~b] are theorems. G.M.}

                          end;
            fusion      : begin
                          r2parts:=r2parts+[x];
                          if ((lsub[x]=rsub[x]) and (contra[lsub[x]]<>0))
                             then add_init(x,contra[lsub[x]])

			     {[a o a, ~a] is a theorem. G.M.}

                             else
                          if ((contra[lsub[x]]<>0) and (contra[rsub[x]]<>0))
                             then begin
                                  add_init(contra[lsub[x]],contra[rsub[x]]);
                                axiom_g[theorem_cnt]:=axiom_g[theorem_cnt]+[x];
                                axiom_r[theorem_cnt][3]:=1;
                                  end;

                             {[~a, ~b, a o b] is a theorem G.M.}

                          end;
            conjunction : inverts:=inverts+[x];
            otherwise   : atparts:=atparts+[x];
            end;  {case}
        strongparts:=strongparts+[bigf,bigt]-[0];
        r2parts:=r2parts+[bigt]-[0];
        end; {pre_tabs.maktab}

    {The local variables and types defined within this function are:

     type    tmpsubs = record s:symtyp; loc,l,r:genrange; c:real end;
     var     x       : integer;
             srt     : array[0..sublim] of tmpsubs;
             realoc  : packed array[0..sublim] of genrange;

     The basic structure of this function is

	if getsub(rootsub) then maktab;

     getsub will fetch the next item from line. It will be then converted
     to appropriate symbol (if possible), and then getsub will
     check if it is an operator. If not, then the corresponding record
     (tmp, declared locally in getsub, to be of type tmpsubs) will be inserted 
     into srt (with insert). The cost of the record will be set to inicost and 
     left and right pointers will be set to NIL. If the symbol IS an operator, 
     then getsub will call itself recursively for left and right sides of the 
     operator (in this order). The cost for the symbol will then be calculated 
     according to the type of the symbol and the costs of its left and right 
     branches. Finally the corresponding record (tmp) will be inserted into 
     srt (with insert).  getsub will return true if the formula is correct 
     and false otherwise.  (It must have all its right and left parts in place 
     to be correct). Its var parameter will be set to the index of tmp in srt.
     NOTE that for this function to work the input formula MUST be given
     in the stack notation with operators first. At the end of this function 
     every token from the formula will have its index in srt and rootsub
     will be set to point to the principal connective.

     maktab will sort entries in srt according to cost and then according to 
     symbol ordering. The .l and .r pointers will be reallocated. Information 
     contained in srt is then transferred to global tables, most notably to 
     sym (symbols corresponding to generators), subcost (cost of the 
     generator), lsub (pointers to left parts of formulae, if any), rsub 
     (pointers to right parts of formulae, if any). Once this is done all 
     specific cases of basic theorems for the variables used in the 
     input formula will be derived and stored on axiom_g and axiom_r. 
     If <a + ~a> pair is encountered then the record of inverts will be 
     appended to. If disjunction is encountered then the record of strongparts 
     is appended to. If fusion is encountered the record of r2parts will be 
     appended to. For other symbols the record of atparts is appended to.

     G.M.}



    begin {pre_tabs}
    pre_tabs:=false;
    purge(root);
    submax:=0;
    getln;
    if getsub(rootsub) and (linloc=linlen[0])
       then begin maktab; pre_tabs:=true; end
       else if linlen[0]<>0 then writeln('Error in formula');
    end; {pre_tabs}


{----- PRINT-UP ROUTINES -----------------------------------------------------}

procedure print_form(s:genrange);

{sym[s] contains symbols (i.e., fission..nonsymbl) corresponding to 
 appropriate value of index s. print_form will convert this symbol to
 its corresponding ascii sequence and will print it on stdout. It will
 then do the same for left and right parts of the formula. I.e. it will
 call print_form(lsub[s]), which upon printing out a token will call
 print_form(lsub[lsub[s]]), and so on, until one of these (lsub[lsub[lsub...)
 is 0. Likewise with the right part of the formula. This suggests, of course,
 that filling lsub and rsub with numbers must be done in a corresponding
 manner to the way they are being written out.
 Called from print_mset - and recursively from itself.
 G.M.}

    begin
    if s<>0
       then begin
            if tok[sym[s]][2]=nul 
	       then write(tok[sym[s]]:1)
	       else write(tok[sym[s]]);
            print_form(lsub[s]);
            print_form(rsub[s]);
            end;
    end; {print_form}

{procedure print_mset(n:node_ptr;prtsb:boolean;dpth:integer); forward; G.M.}

procedure print_mset;

{This procedure will print the part of the tree with its root at
 node_ptr marking the depth of the node with indentation. print_form
 is called to convert the number of the generator into the corresponding
 formula. The number of repetitions is then printed after ':'. The
 Principal Constituent (the pc) is additionally marked by '*'.
 If the state of the node is 'theorem', the line is marked with the
 word 'CLOSED'. If prtsb (print subtree?) is true and if the main connective 
 of the principal constituent is conjunction or fusion then the procedure 
 will recursively descend down the tree (left first) and print the multisets
 it encounters until a multiset without fusion or conjunction is found.
 Called from error, main, and add_baddy.
 G.M.}

    var     s     : node_ptr;
            gr    : genray;
            gs    : genset;
            x     : genrange;
    begin
    if n<>0
       then with tree[n] do
            begin
            gs:=gens; 
	       {tree[n].gens is the set of generators of the multiset. G.M.}
            trans(gs,gr);
	       {put indexes (ord - really) of generators from gs to gr. G.M.}
            for s:=1 to dpth do write('  ');
            x:=1;
            while (x<=repmax) do
               if gr[x]>0
                  then begin
                    print_form(gr[x]);
                    write(':',reps[x]:(reps[x] div 10)+1);
                    if gr[x]=pc then write('*  ') else write('  ');
                    x:=x+1
                    end
                  else x:=(repmax+1);
            if prtsb
               then begin
                    case state of
                      theorem : if lp=0 then write('CLOSED');
                      otherwise: write(state)
                    end;
                    writeln;
                    case sym[pc] of
                      conjunction,
                      fusion      : begin
                                    s:=lp; print_mset(s,prtsb,dpth+1);
                                    s:=rp; print_mset(s,prtsb,dpth+1);
                                    end;
                      otherwise   : begin
                                    s:=lp; print_mset(s,prtsb,dpth);
                                    end;
                    end
                    end
               else writeln;
            end;
    end; {print_mset}




{----- INTERUPT ROUNTINES ----------------------------------------------------}

procedure inform;

{This procedure writes on stdout information about timing, number of
 nodes in the tree, general state of the formula (ACTIVE, PROVED, REFUTED),
 depth, width and number of open paths for the proof search tree.
 Called from: info, try_proof, error, and main.
 G.M.}

    var     x : integer;
    begin
    time(nam); write(nam);
    case tree[root].state of
        initial,
        open    : x:=(clock-strtime) div 1000;
        failed  : if root=0 then x:=0 else x:=endtime;
        theorem : x:=endtime;
    end;
    write('after ',xp(x),'secs.');
    x:=node_lim-node_cnt;
    write('  Nodes=',xp(x),' State:');
    if root=0
       then writeln(' UNDEFINED')
       else with tree[root] do
                case state of
                  open : writeln(' ACTIVE');
                theorem: writeln(' PROVED');
                failed : writeln(' REFUTED');
                end; {case}
    write('   Depth is ',xp(count[depth]), ' max:',xp(count[maxdpth]));
    write('   Width is ',xp(count[width]), ' max:',xp(count[maxwdth]));
    write('   Open paths ',xp(count[srch]));
    writeln;
    end; {inform}


procedure info;

{This procedure will in the first place call inform. Then it will write on 
 standard output input line and the comment, number of scans of the pst that 
 took place during the proof, number of constructed, reduced, failed and 
 closed nodes. It will then specify how many nodes failed due to Curry, 
 Pos/Neg, Rule2, and matrices. Note that neither info nor inform will
 print the proof. The proof is printed by print_mset which is called
 separately from main.
 Called from try_proof, error, and main.
 G.M.}

    var     fc,sc,x,y : integer;
            xc        : array[count_typ] of integer;
    begin
    inform;
    writeln;
    write('Info for :  ');
    for y:=1 to linlen[0] do write(line[0,y]); writeln;
    write('Comment  :  ');
    for y:=1 to linlen[1] do write(line[1,y]); writeln;
    writeln;
    writeln('    ',xp(count[scans]),'SCANS occurred.');
    for cc:=scans to depth do xc[cc]:=0;
    fc:=0;
    for cc:=curried to mat do fc:=fc+count[cc];
    for cc:=curried to mat do xc[cc]:=(count[cc]*100) div max(fc,1);
    sc:=count[aclosed]+count[dclosed];
    x:=sc;
    y:=(count[dclosed]*100) div max(sc,1);
    writeln('    ','Of ',xp(count[made]),'nodes constructed: ');
writeln(ht,xp(count[idemred]),'reduced, ',xp(fc),'failed & ',xp(x),'closed');
    writeln(ht,   '% Failures: Curry=',xp(xc[curried]),
                               ' Pos/Neg=',xp(xc[swparts]),
                                 ' Rule2=',xp(xc[rul2]),
                              ' Matrices=',xp(xc[mat]));
    write(ht,    'Derived closures= ',xp(y),'%');
    write('   Watching ',xp(baddy_cnt));
    writeln('  Suspensions=', xp(count[susd]));
    writeln;
    end; {info}



{----- PROVER ROUTINES -------------------------------------------------------}

procedure try_proof; {try_proof}
    var     tmp : packed array[0..sublim] of genrange;

    function matrix_test(n:node_ptr) : boolean; {try_proof.matrix_test}
        label   40,41;
        var  numvar,
                x,y : integer;
               valt : array[var1..var5] of integer;
              atray : array[1..5] of var1..var5;
               varn : var1..var5;
               pset : partset;
               tloc : reprange;
               trep : repray;
               gray : genray;

        function eval(s:genrange) : integer; {try_proof.matrix_test.eval}

	{This function returns the logical value for a formula with its
	 root given by s. The variables are evaluated to the corresponding
	 values passed in valt[sym[s]]. The evaluation then proceeds using
	 data stored in contab, distab, fustab, and fistab.
	 Called from f_eval only.
	 G.M.}

            begin
            with matrix do
            case sym[s] of
              conjunction : eval:=contab[eval(lsub[s]),eval(rsub[s])];
              disjunction : eval:=distab[eval(lsub[s]),eval(rsub[s])];
              fusion      : eval:=fustab[eval(lsub[s]),eval(rsub[s])];
              fission     : eval:=fistab[eval(lsub[s]),eval(rsub[s])];
              big_t       : eval:=siz;
              big_f       : eval:=0;
              sml_t       : eval:=ldes;
              sml_f       : eval:=negtab[ldes];
              otherwise   : if sym[s] in [var1..var5]
                               then eval:=valt[sym[s]]
                               else eval:=negtab[valt[dual[sym[s]]]];
              end;
            end; {try_proof.matrix_test.eval}

        function f_eval : boolean; {try_proof.matrix_test.f_eval}

	{This function accelerates the evaluation of a formula against
	 the matrix. It uses a multiple fission table to do multiple
	 fissions, and calls eval on subformulas. See the comment
	 within the body of f_eval for more information. F_eval returns
	 TRUE if a formula is valid and false otherwise. Note that
	 the values substituted in place of variables are still taken
	 from valt by eval. So, this formula does the testing only for
	 the particular set of values in valt.
	 Called from try_proof.matrix_test.f_eval only
	 G.M.}

            var x,y : integer;
            begin
            with matrix do
                begin

		{Set y to the negation of the lowest designated value. 
		 G.M.}

                y:=negtab[ldes]; 

		{tloc will be set to tree[n].locmax by try_proof.matrix_test.
		 The loop below goes through all the generators of the
		 formula in the node. The generators are separated by
		 fission operators in the formula (i.e., the commas) - 
		 for every generator its corresponding logical value is
		 found by eval and then idfis is used to evaluate what
		 this value will come up to be if that generator is
		 repeated trep[x] times in the formula. A fission of
		 that value with the lowest value of y so far is then
		 found and put back into y. If at the end of the day
		 y is within the range of designated values f_eval returns
		 TRUE otherwise it returns FALSE.
		 G.M.}

                for x:=1 to tloc do
                    y:=fistab[y,idfis[eval(gray[x]),min(siz,trep[x])]];
                f_eval:=(y in des);
                end;
            end; {try_proof.matrix_test.f_eval}

	{This function does the obvious thing: it assigns all possible
	 values to the variables of the formula found at the node of the
	 tree and checks if combining those with truth tables for all
	 the operators used in the formula will return less than ldes
	 for the whole formula. If that happens further tests are stopped
	 and FALSE is returned. If that does not happen, TRUE is returned.
	 Called from try_proof.scan.filter only.
	 G.M.}

        begin {try_proof.matrix_test}

		{Convert the set of generators to the array of generators. 
		 G.M.}

                trans(tree[n].gens,gray);

		{Create a set of all parts in the formula. G.M.}

                pset:=[];
                tloc:=tree[n].locmax;
                trep:=tree[n].reps;
                for x:=1 to tloc do
                    pset:=pset+posp[gray[x]]+negp[gray[x]];
                numvar:=0;

		{Initialize all variables in the formula to 0.
		 G.M.}

                for varn:=var1 to var5 do
                    if varn in pset
                       then begin
                            valt[varn]:=0;
                            numvar:=numvar+1;
                            atray[numvar]:=varn;
                            end;

		{Perform the matrix test against the assigned values
		 for the variables. If the test is positive, increment
		 the value of the "next" variable (note the nasty goto
		 statement inside the for x:=1 to numbar loop!) and
		 repeat the test. If the test is negative return FALSE
		 immediately. Only after all the tests for all possible
		 assignments return TRUE, TRUE is returns by matrix_test
		 itself.
		 G.M.}

                with matrix do
                        if working
                           then begin
                            40: if f_eval
                                   then for x:=1 to numvar do
                                     if valt[atray[x]]<siz
                                        then begin
                                             valt[atray[x]]:=valt[atray[x]]+1;
                                             goto 40
                                             end
                                        else valt[atray[x]]:=0
                                   else begin
                                        count[mat]:=count[mat]+1;
                                        matrix_test:=false;
                                        goto 41
                                        end;
                                end;
                matrix_test:=true;
   41:  end; {try_proof.matrix_test}


function scan(n:node_ptr) : node_state; {try_proof.scan}
    var     lson,rson   : node_ptr;
            gray        : genray;


    function part_chk : boolean; {try_proof.scan.part_chk}

    {In order to understand what is happening inside this function
     some definitions and theorems from Kama Sutra (pp. 66-69) should
     be reviewed:
	1) parts: set of literal subformulas - can be positive (+) and
		  negative (-);
	2) strong +/- property: every variable is both in positive and
		  negative parts;
	3) weak +/- property: there are some variables that are both in
		  positive and negative parts OR T is in positive parts;
	4) strict +/- property: explicit propositional variables must have
		  negated counterparts which are members of complex 
		  subformulas;
	5) If no disjunctions, or T or F in parts then there must be
		  strong +/- for L5 provability (occurrences of disjunctions
		  are ticked on strongparts);
	   else
		  there must be weak +/- for L5 provability;
	6) If no constants then there must be strict +/- for L5 provability;
	7) rule-of-2: no fusion, T, or explicit f, then cardinality of
		  subformula must be less than 3 for L5 provability (fusions
		  are ticked on r2parts);

     This function will first produce sets of positive and negative
     single atoms and positive and negative complex subformulas, it will
     also find the cardinality of the formula. It will then test the formula
     against strong +/- parts and rule of 2, if applicable; weak +/- parts
     and rule of 2, if applicable; and finally if all the previous tests
     are passed, it will check against the strict +/- parts.

     Called from try_proof.scan.filter only.
     G.M.}

        var       x,y : genrange;
            atpos,		{positive atoms, G.M.}
            atneg,		{negative atoms, G.M.}
            cpos,cneg,		{positive and negative connectives;
				 they are really positive and negative
				 complex subformulas, but they will be
				 called connectives for short, G.M.}
            poss,negs : partset;   {positive and negative connectives and
				    atoms, G.M.}
            rtyp      : genset;	   {all subformulas of a multiset, G.M.}
            crd       : integer;   {cardinality of a multiset, G.M.}
            ptret     : boolean;   {value to be returned from the +/- check,
				    G.M.}
        begin
                ptret:=false;	{be pessimistic, and initialize all sets
				 and counters to [] or 0, G.M.}

                poss:=[]; negs:=[]; rtyp:=[]; crd:=0;
                atpos:=[]; atneg:=[]; cpos:=[]; cneg:=[];
                with tree[n] do
                    begin

		    {Go through all the generators of the multiset
		     in the node of the tree. G.M.}

                    for x:=1 to locmax do
                        begin

			{Evaluate the cardinality of the multiset. G.M.}

                        crd:=crd+reps[x];
                        if gray[x] in atparts

			{Fill up the atom parts. G.M.}

                           then begin
                                atpos:=atpos+posp[gray[x]];
                                atneg:=atneg+negp[gray[x]];
                                end
                           else begin

			{Fill up the connectives parts. G.M.}

                                cpos:=cpos+posp[gray[x]];
                                cneg:=cneg+negp[gray[x]];
                                end;

			{Fill up the set of all parts. G.M.}

                        rtyp:=rtyp+subs[gray[x]];
                        end;
                    poss:=atpos+cpos; {all positive parts. G.M.}
                    negs:=atneg+cneg; {all negative parts. G.M.}

		    {if there are no disjunctions, F, and T...  G.M.}

                    if (rtyp*strongparts)=[]
		      
		       {then check for strong +/- property. G.M.}

                       then if poss=negs

			       {if passed that then check if fusions,
				T, or f are present... G.M.}

                               then if (rtyp*r2parts)=[]

				       {if they aren't then check the
					cardinality, i.e. the rule-of-2. G.M.}

                                       then if crd<3
                                               then ptret   :=true
                                               else count[rul2]:=count[rul2]+1
                                       else ptret   :=true
                               else count[swparts]:=count[swparts]+1

		    {Variable count merely gathers the statistics for
		     display by info.
		     G.M.}

	            {Now, there might be some disjunctions, F, or T, in the
		     formula. In this case check for the weak +/- property... 
		     G.M.}

                       else if (big_t in poss) or ((poss*negs)<>[])
			       
			       {if passed that then repeat the rule-of-2
				check. G.M.}

                               then if (rtyp*r2parts)=[]
                                       then if crd<3
                                               then ptret   :=true
                                               else count[rul2]:=count[rul2]+1
                                       else ptret   :=true
                               else count[swparts]:=count[swparts]+1;
		    
		    {Finally check for the strict +/- property. First,
		     are there any complex formulas? G.M.}

                    if ptret and ((cpos+cneg)<>[])
		       
		       {if so, then check if positive (negative) atoms
			have counterparts in negative (positive) subformulas.
			G.M.}

                       then ptret:=((atpos*cneg)=atpos) and
                                   ((atneg*cpos)=atneg);

                    end; {with tree do... G.M.}
        part_chk:=ptret;
        end; {try_proof.scan.part_chk}



    function filter : node_state; {try_proof.scan.filter}

    {This function applies first part_check and if that is passed,
     it applies matrix_test. The node is marked as open if both
     tests are successful, otherwise as failed.
     Called from try_proof.scan only
     G.M.}

        begin
        filter:=failed;

	{Note that part_chk knows about the node it should work on,
	 because scan is called by try_proof with the particular node
	 passed as an argument. This node is then accessed by part_chk
	 as a variable (n) whose scope is global within scan.
	 G.M.}

        if part_chk
           then if matrix_test(n)
                   then filter:=open
        end; {try_proof.scan.filter}

    function curry_fail : boolean; {try_proof.scan.curry_fail}

    {This function tests the whole branch terminated with the given node
     against Curry property. Curry property means the following: if
     a multiset obtained by transforming a given multiset according to
     the transformation rules is still cognate with the given multiset, then 
     it should be simpler for at least one of the generators, i.e., at least 
     one of them should have reduced repetitions. 
     Called from try_proof.scan only.
     G.M.}

        label   88;
        var x     : integer;
            predec: node_ptr;

        begin
        curry_fail:=false;
        predec:=tree[n].up;
        with tree[n] do
        while predec<>0 do
	    
	    {Check if the multisets at the given node and at the predecessor
	     node are cognate. G.M.}

            if gens=tree[predec].gens
               then begin
		    
		    {if they are cognate, then check if at least some
		     simplification took place while moving from the
		     predecessor to the given node. G.M.}

                    for x:=1 to card(gens) do if tree[predec].reps[x]>reps[x]
                            then goto 88;

		    {if no simplification was obtained then this branch
		     is clearly bad. Such branch is said to violate Curry 
		     property and must be pruned.
		     See Kama Sutra, def 2.7, p. 34, and theorem 2.8, p.
		     35. G.M.}

                    count[curried]:=count[curried]+1;
                    curry_fail:=true; predec:=root;

		    {if the program jumped here from the goto above
		     then move one floor up above the predecessor,
		     and compare the node there against the given node.
		     G.M.}

                    88: predec:=tree[predec].up;
                    end
               else predec:=tree[predec].up;
        end; {try_proof.scan.curry_fail}


    function not_baddy : boolean; {try_proof.scan.not_baddy}

    {This function will go through the list of bad multisets. If
     a cognate match is found, then the repetitions will be compared
     for each of the generators. If there is at least one such generator
     in the tested multiset that the number of repetitions for that
     generator is less than the number of the repetitions in the "bad"
     multiset, the tested multiset is considered "not bad". Hence, only
     a multiset which strongly contains the bad multiset will be marked
     as being also bad. In this case the state slot of the tree node is
     marked FAILED, and the function returns FALSE.
     Called from try_proof.scan only.
     G.M.}

        var     x : integer;
               tr : repray;
               tb : boolean;
        begin
        not_baddy  :=true; {Be optimistic. G.M.}
        with tree[n] do
          begin
          if failcnt>=0 then x:=0
                        else 

			     {Note that this clause will be executed
			      only if failcnt is negative. In this case
			      The loop below will not be executed,
			      because x will be set to baddy_cnt. 
			      Note also that failcnt is never used
			      outside of this function. If you then look
			      closely at how it is used within this
			      function, you will notice that it is
			      not used at all! Most Pascals would initialize
			      failcnt to 0 at the beginning of the 
			      execution. So that this close would not
			      be executed at startup. For a node, not
			      marked as bad, failcnt will remain 0. 
			      If a node is found to be bad, failcnt will
			      jump to 50. Because failcnt IS NOT referred
			      to anywhere else, it seems that it is a spurious
			      variable.
			      G.M.}

			begin x:=baddy_cnt; failcnt:=failcnt+1 end;
          while x<baddy_cnt do
            begin
            x:=x+1;

	    {If the multiset at the node of the tree is cognate with
	     one of the bad multisets ... G.M.}

            if bad_g[x]=gens
               then begin
                    tb:=true;

		    {If there is at least one generator in the tree
		     multiset such that it repeats less times than the
		     corresponding generator in the bad-guy-list multiset
		     then the node is still not bad... G.M.}

                    for y:=1 to card(gens) do if (reps[y]<bad_r[x][y])
                        then tb:=false;
                    if tb

		    {... otherwise the node is bad. In this case
		     return FALSE, mark the node state as FAILED,
		     set up failcnt to 50 (is the the maximum number
		     failcnt is allowed to grow to?), and set
		     x to baddy_cnt, so that the loop (while x<baddy_cnt)
		     will be aborted. I'd rather use plain "goto" in this
		     case!G.M.}

                       then begin
                            not_baddy:=false;
                            state:=failed;
                            count[susd]:=count[susd]+1;
                            failcnt:=50;
                            x:=baddy_cnt;
                            end;
                    end;
            end;
          end;
        end; {try_proof.scan.not_baddy}

    function chk_theorem : boolean; {try_proof.scan.chk_theorem}

    {This function will test the multiset contained at the node of the
     tree against all axioms stored in the axiom list. If the tested
     multiset uses the same generators as an axiom (and possibly some
     other ones), then idempotence reductions are carried out in the
     first place. The generators _not_ used in the axiom are squashed
     to 1 occurrence only, whereas the generators which have been
     used in the axiom are squashed to 2 occurrences if the number
     of their occurrences is larger than 2. If the multiset is cognate
     with an axiom, then we test if it is strongly contained in the axiom.
     If that is the case the match is found and chk_theorem returns TRUE.
     Also various global counters are maintained.
     Called from try_proof.scan only.
     G.M.}

        var     x : integer;
               tr : repray;
               tb : boolean;
        begin
        chk_theorem:=false;  x:=0; {be pessimistic. G.M.}
        with tree[n] do
          begin

	  {if "T" is amongst the generators then the multiset is
	   a tautology, and that's where the processing ends. G.M.}

          if bigt in gens
             then begin
                  count[aclosed]:=count[aclosed]+1;
                  chk_theorem:=true; x:=theorem_cnt;

		  {Note that again x:=theorem_cnt instead of "goto"
		   or some other more obvious device to not to
		   do the while loop below. G.M.}

                  end;
          tr:=reps;
          while x<theorem_cnt do
            begin
            x:=x+1; tb:=false;
            if (x<=axiom_cnt)

	       {If all the generators used in the axiom are also
		present in the tested multiset then carry out the
		idempotence reductions ... 
		G.M.}

               then if (axiom_g[x]*gens)=axiom_g[x]

		       {Go through all the generators in the
			multiset ... G.M.}

                       then for y:=1 to locmax do

				{... and if a given generator is not
				 used in the axiom... G.M.}

                                if not(gray[y] in axiom_g[x])
                                   then begin

					{... then squash its corresponding
					 repetition number to 1... G.M.}

                                        if reps[y]>1
                                           then begin
                                                reps[y]:=1;
                                                tb:=true;
                                                end
                                        end

				   {... in turn if it has been used in 
				    the axiom then squash its corresponding
				    repetition number to 2 if its larger than
				    2... G.M.}

                                   else if reps[y]>2
                                           then begin
                                                reps[y]:=2;
                                                tb:=true;
                                                end;

	    {...whenever some squashing took place, increment the
	     idempotence reduction counter... G.M.}

            if (tb=true) then count[idemred]:=count[idemred]+1;

	    {Now if the generators in the given multiset are exactly the
	     same as the generators used in the axiom ... G.M.}

            if axiom_g[x]=gens

	       {... then if there is at least one generator in the tested 
		multiset which repeats more times than the corresponding 
		generator in the axiom, the tested multiset is not a theorem. 
		In other words, for the given multiset to be marked a theorem 
		it must be strongly contained in the axiom.
		G.M.}

               then begin
                    tb:=true;
                    for y:=1 to card(gens) do if (reps[y]>axiom_r[x][y])
                        then tb:=false;
                    if tb

		    {At this stage we already know that the tested multiset
		     is a theorem, so we set chk_theorem to TRUE and we
		     increment the corresponding counters. G.M.}

                       then begin
                            chk_theorem:=true;
                            if x<=axiom_cnt
                               then count[aclosed]:=count[aclosed]+1
                               else count[dclosed]:=count[dclosed]+1;
                            x:=theorem_cnt;
                            end;
                    end;
            end;
          end;
        end; {try_proof.scan.chk_theorem}


    function getpc : node_state; {try_proof.scan.getpc}

    {Get principal constituent. This function calls frstpc, which
     fetches next generator from the list of generators of the
     multiset which lives at the node of the tree. The current generator
     lives in tree.pc. getpc then checks if the new generator is
     either disjunction or fusion. If it is any of those then
     getpc proudly returns OPEN thus opening a new branch of the
     tree, otherwise it returns FAILED. 
     Called from try_proof.scan.apply_fusion and from try_proof.scan only.
     G.M.}

        var     gs  : genset;
                x   : genrange;

        begin

	{lson and rson are internal globals of try_proof.scan. They are
	 node pointers and are set up within the body of scan itself.
	 G.M.}

	{Whatever lson and rson point to at this stage is purged. G.M.}

        purge(lson); purge(rson);
        with tree[n] do
            begin
            gs:=gens; x:=pc;

	    {Find the lowest generator in gs that is larger than x and
	     put it in x if it exists. frstpc returns FALSE if the
	     generator is not found. G.M.}

            if frstpc(gs,x)

	       {At this stage we have a new generator sitting in x.
		If it's either disjunction or fusion, getpc returns
		OPEN - presumably a new branch of the proof search
		tree is opened in this way. Otherwise getpc returns
		FAILED. FAILED is also returned if frstpc couldn't
		find a generator larger than x in gens, in other
		words if x is the last generator.
		G.M.}

               then if sym[x] in [disjunction,fusion]
                   then begin pc:=x; getpc:=open; pcloc:=pcloc+1; stage:=0 end
                   else getpc:=failed
               else getpc:=failed;
            end;
        end; {try_proof.scan.getpc}

    procedure defson(var son:node_ptr); {try_proof.scan.defson}

    {This procedure defines the new multiset at the new node of the
     tree. The information about the multiset to be transferred to
     that node lives on tmp[x]. tmp[x] is an array, with x going through
     all the generators of the original formula (x:=1 to submax; submax
     being set in pre_tabs.getsub.insert). If a given generator x is used 
     in the new formula n times then tmp[x] = n. If tmp[x] = 0 then it is
     understood that the generator x is not used in the new formula.
     (tmp is initialized within the body of try_proof.scan.create, which
     will in turn call defson. try_proof.scan.create will also malloc
     a new node and will put the pointer to it into son.)
     The transfer then is carried out via: reps[y] := tmp[x] and
     gens := gens + [x]. If "f" is encountered in tmp[x], its corresponding
     tmp is automatically set to 0. In other words we ignore all "f"s.
     Various other node counters such as state, up, pcloc, locmax
     and stage are for the new node are also initialized by this
     procedure.
     Called from try_proof.scan.create, try_proof.scan.apply_fusion.getright,
     and try_proof.scan.apply_fusion.getleft.
     G.M.}

        var     x,y : genrange;
        begin
        y:=0;
        with tree[son] do
            begin
            tmp[smlf]:=0; {smlf defines the position of "f" in 
			   the set of generators (the set is ordered). 
			   if "f" occurs on its own amongst the generators
			   we knock it off from the outset. G.M.}

            for x:=1 to submax do 

	    {go through all the generators in the _root_ formula and check
	     if they occur also in the given formula, which is represented
	     here by tmp. tmp[x] is a number of times generator x will
	     occur in the multiset, which will be installed at this node.
	     If a generator is flagged in tmp, then add it to gens and
	     copy the number from tmp to reps. In this way the multiset
	     at the new node is set up.
	     G.M.}

                if tmp[x]<>0
                   then begin
                        y:=y+1;
                        if y=1 then pc:=x;
                        reps[y]:=tmp[x];
                        gens:=gens+[x];
                        end;
            state:=initial; up:=n; pcloc:=1; locmax:=y; stage:=0;
            end;
        end; {try_proof.scan.defson}


    procedure create(var son:node_ptr; ad1,ad2:genrange);
    {try_proof.scan.create}

    {This procedure will copy the multiset from the present node of the
     tree into tmp. It will then modify it (--tmp[pc], ++tmp[ad1], ++tmp[ad2])
     and call defson on a newly malloced node pointed to by son. Defson will
     throw away every single occurrence of "f" and will copy tmp onto
     the multiset at the new node, whose other parameters (state, up, pcloc,
     locmax, and stage) will be also fully initialized. 
     Called from try_proof.scan only
     G.M.}

        var     x,y : genrange;
        begin

	{malloc a new node and put the pointer to it into son. G.M.}

        assign(son); y:=0;
        with tree[n] do
            begin

	    {go through all generators in the root formula. G.M.}

            for x:=0 to submax do

		{if a generator is found in the formula that resides at
		 the investigated node then transfer the corresponding
		 number of repetitions of that generator to tmp [x]. G.M.}

                if x in gens
                   then begin y:=y+1; tmp[x]:=reps[y]; end

		   {...otherwise put 0 in tmp[x]. Compare this with what
		    happens in defson above! G.M.}

                   else tmp[x]:=0;

	    {We must assume that at this stage KRIPKE will know what
	     pc is and what ad1 and ad2 are. Number of counts that 
	     correspond to the principal constituent (pc?) is reduced,
	     and number of counts corresponding to ad1 and ad2 are
	     incremented. G.M.}

            tmp[pc]:=tmp[pc]-1;   {Create is called to fork a new branch
				   of the tree. In this process the main
				   connective is thrown away, that's why
				   tmp[pc]:=tmp[pc]-1. G.M.}

            tmp[ad1]:=tmp[ad1]+1; {This corresponds to what the left hand
				   side of the main connective was. It
				   was not a generator in its own right,
				   but it will be after the forking,
				   that's why tmp[ad1]:=tmp[ad1]+1. G.M.}

            tmp[ad2]:=tmp[ad2]+1; {This corresponds to what the right hand
				   side of the main connective was. Ditto. 
				   G.M.}
            end;

	{Now we've done our job with setting up tmp, and we can pass it
	 to defson, which will transfer it onto a new multiset at the
	 new node of the tree. Note that if not for the operations on the 
	 tmp[pc], tmp[ad1], and tmp[ad2], we would end up transferring the 
	 multiset from the old node to the new one without change. G.M.}

        defson(son);
        end; {try_proof.scan.create}


    procedure apply_fusion; {try_proof.scan.apply_fusion}

        var tr : repray;

        function getright : node_state; {try_proof.scan.apply_fusion.getright}
            var     x   : genrange;
                acting  : boolean;

	    function eql : boolean; {try_proof.scan.apply_fusion.getright.eql}

	    {This function will check if the multiset on the left hand
	     side of the principal constituent is the same as the multiset
	     on the right hand side of the principal constituent. 
	     Called from try_proof.scan.apply_fusion.getright.eql only.
	     G.M.}

		var	x : genrange;
		begin
		eql:=false;

		{If the sets of generators on both sides are the same... G.M.}
		{lson and rson are globals within the body of scan. G.M.}

		if tree[lson].gens=tree[rson].gens
		   then begin
			eql:=true;

			{... then check if the repetition numbers are the same
			 too. G.M.}

			for x:=1 to repmax do
			if tree[lson].reps[x]<>tree[rson].reps[x]
				   then eql:=false;
			end;
		end; {try_proof.scan.apply_fusion.getright.eql}

	    {It is not quite clear what this function does, but it seems
	     that it is doing the following: 
		Check if rkey at the present node of the tree is set up.
		If it is then it means that a new branch of the tree
		corresponding to the right hand side of the formula
		should be forked, and consequently a new node of the
		tree is allocated. A temporary array, tmp[x], is then
		initialized and then the contents of rmsk is transferred
		to it. Additional increment operation is done on the entry
		of tmp corresponding to pc. After this "defson" is called
		which will transfer the stuff from tmp[x] to the multiset
		at the new node and which will initialize other parameters
		at that node. Comparison is then made between the multiset
		at the left node and the newly created multiset at the right
		node (which means that the left node must exist by now!). 
		If they are identical the right node is declared
		LINKED. Once this is done an operation takes place which
		probably results in removing information referring to the
		left node (lmsk) from the information referring to the right
		node (rmsk). After all this is finished getright will
		return INITIAL. 
		If, however rkey has not been set up for creation of the
		new branch then getright will return FAILED.
	     The details of what's going on may become clearer when
	     this procedure is looked back on from higher level procedures
	     such as scan and try_proof.
	     Called from try_proof.scan.apply_fusion.check_fusions,
	     and try_proof.scan.apply_fusion.
	     G.M.}

            begin {try_proof.scan.apply_fusion.getright}

	    {tree[n].rkey flags that the right son of the node (rson)
	     should be created. If that flag is up (i.e. > 0) then... G.M.}

            if tree[n].rkey>0
               then begin

		    {... allocate space for the new node and return a pointer
		     to it in rson... G.M.}

                    assign(rson);
                    with tree[n] do
                        begin

			{Zero number of occurrences in tmp for each generator 
			 used in the root formula. Or, in other words, 
			 reinitialize tmp[x]. G.M.}

                        for x:=0 to submax do tmp[x]:=0;

			{Generator array, gray, is an array which corresponds
			 to the generator set. This array is made by trans.
			 Its last entry is tree[n].locmax. I think (for the 
			 time being)
			 that within the body of scan or try_proof, gray
			 contains the elements of the multiset on which
			 work is being done at the moment. This gray is a
			 global within the body of scan, anyway. Not quite sure
			 what tree[n].rmsk[x] is - sounds and looks like 
			 right-son's-mask, it is a part of the node which
			 is of the same type as the repetition array reps. 
			 Only the generators that are flagged within rmsk 
			 would be filtered through to tmp at this stage. G.M.}

                        for x:=1 to locmax do tmp[gray[x]]:=rmsk[x];

			{If there is anything that should be marked as
			 right-hand-side's principal connective, the 
			 corresponding entry of tmp is incremented -
			 this might be how the marking is done within
			 tmp. But note that within the body of
			 defson, which follows, pc will be set to 
			 the first generator x in tmp, for which
			 tmp[x] <> 0. G.M.}

                        tmp[rsub[pc]]:=tmp[rsub[pc]]+1;

			{Now copy tmp onto the multiset at the new node
			 rson. Initialize other parameters at that node.
			 G.M.}

                        defson(rson);

			{Check if the multisets at rson and lson are
			 identical and if they are then declare the state
			 of the rson-node as LINKED. G.M.}

                        if eql
                           then tree[rson].state:=linked;
                        acting:=true;
                        while acting do
                            begin

			    {tr is the repetition array defined only
			     within the body of apply_fusion. G.M.}

                            if (tr[rkey]=1) and (rmsk[rkey]=0)
                               then begin

			       {Now, this piece of code is not immediately
				obvious. It looks like we are modifying
				the rmsk so as to get rid of the guys who
				are already on the left hand side of the
				formula: "rmsk[x]:=tr[x]-lmsk[x]". This loop
				is being executed for tree[n].rkey = 
				tree[n].locmax and rkey is being decremented 
				at every step until the conditions at the top 
				or rkey=0 are hit, at which stage the loop is
				aborted through "acting=false". G.M.}

                                    acting:=false;
                                    rmsk[rkey]:=1;
                                    for x:=(rkey+1) to locmax do
                                           rmsk[x]:=tr[x]-lmsk[x];
                                    rkey:=locmax;
                                    end
                               else rkey:=rkey-1;
                            if rkey=0 then acting:=false;
                            end;
                        getright:=initial;
                        end;
                    end

	       {This option is taken if tree[n].rkey <= 0. G.M.}

               else getright:=failed;
            end; {try_proof.scan.apply_fusion.getright}




        function getleft : node_state; {try_proof.scan.apply_fusion.getleft}
            var     x   : genrange;
                finished: boolean;

          procedure getl; {try_proof.scan.apply_fusion.getleft.getl}

	  {This little function does to lson much of what getright does
	   to the rson. And so, it initializes tmp[x] and then it filters
	   stuff into it through lmsk. The entry of tmp which corresponds
	   to pc is incremented. At this stage a new node is allocated for
	   lson and the contents of tmp are trasferred from tmp to the multiset
	   at the new node by defson. The rmsk is modified in a way very
	   similar to what getright does to it too, but the loop goes from
	   1 to locmax this time instead of from (rkey+1) to locmax as in
	   the case of getright. At the end of this process rkey is again
	   set to locmax, and this is again the same as in getright for
	   case ((tr[rkey]=1) and (rmsk[rkey]=0)).
	   Called from try_proof.scan.apply_fusion.getleft only.
	   G.M.}

            begin
            with tree[n] do
                begin
                lkey:=locmax; getleft:=initial; finished:=true;
                for x:=0 to submax do tmp[x]:=0;
                for x:=1 to locmax do tmp[gray[x]]:=lmsk[x];
                tmp[lsub[pc]]:=tmp[lsub[pc]]+1;
                assign(lson);
                defson(lson);
                for x:=1 to locmax do rmsk[x]:=tr[x]-lmsk[x];
                rkey:=locmax;
                end;
              end; {try_proof.scan.apply_fusion.getleft.getl}

	    {This function is rather hard to crack. 
	     It enters a somewhat convoluted loop in the course of
	     which it sets lmsk and lkey, and finally it calls getl 
	     (which will effectively fork a new branch). It returns
	     either FAILED or INITIAL (the latter through getl only).
	     Called from try_proof.scan.apply_fusion.check_fusions,
	     and from try_proof.scan.apply_fusion only.
	     G.M.}

            begin {try_proof.scan.apply_fusion.getleft}
            getleft:=failed;
            finished:=false;
            with tree[n] do
                begin

		{Remove the right branch hanging off this node. G.M.}

                purge(rson);
                repeat
                    case stage of
                     0 : begin

			 {Set up the mask and the flag (lkey) for the
			  left branch and call getl which will malloc
			  a new node and put the right stuff into it.
			  getl will also modify rmsk by subtracting
			  from tr the generators which went to lson.
			  After this clause is finished go to the
			  "otherwise:" clause with stage = 1.
			  G.M.}

                         lmsk:=tree[0].lmsk; lkey:=locmax; getl; stage:=1;
                         end;
                     2 : if reps[pcloc]=1

			 {Even more of the mysterious hocus-pocus than in
			  the "otherwise:" clause. But a chance is now
			  given to quit the REPEAT loop. That happens
			  if reps[pcloc] <> 1.
			  G.M.}

                            then begin
                                 lmsk:=tree[0].lmsk; lkey:=locmax; tr:=reps;
                                 getl; stage:=3;
                                 end
                            else finished:=true;

                     4 : {Nirvana! we can go home! G.M.}
		     
			  finished:=true;

                     otherwise:

		     {This clause does some mysterious hocus-pocus
		      with left mask (note that we are still at the
		      same node!) and left key and then calls getl
		      again. However, stage is an external variable,
		      and it may be that stage=1 will never be
		      reached immediately after stage=0. Only when
		      lkey is reduced to 0, stage is incremented.
		      G.M.}

                        begin
                        while not(finished or (lkey=0)) do
                            if lmsk[lkey] < tr[lkey]
                               then begin
                                    lmsk[lkey]:=lmsk[lkey]+1;
                                    for x:=(lkey+1) to locmax do lmsk[x]:=0;
                                    getl;
                                    end
                               else lkey:=lkey-1;
                        if lkey=0 then stage:=stage+1;
                        end;
                    end; {case}
                until finished;
                end;
           end; {try_proof.scan.apply_fusion.getleft}



        procedure check_fusions; {try_proof.scan.apply_fusion.check_fusions}
        var x,y,z : integer;
        begin
        x:=0;
        with tree[n] do
            begin
            if reps[pcloc]=1 then y:=4
               else if reps[pcloc]=2 then y:=3
                    else y:=reps[pcloc];
            for z:=1 to locmax do
                if z<>pcloc
                   then if reps[z]=1 then y:=y*3
                                     else y:=y*(reps[z]+1);
            tr:=reps; tr[pcloc]:=tr[pcloc]-1;
	    purge(lson);
	    state:=getleft;
	    while (state=initial) do
		begin
		purge(rson);
		state:=getright;
		while (state=initial) do
		    begin
		    x:=x+1;
		    purge(rson);
		    state:=getright;
		    end;
		purge(lson);
		state:=getleft;
		end;
            purge(lson);
            if x<>y
              then begin
                write(n:4,' FAILS fusion_check: pc=',xp(pcloc));
                writeln('pred=',xp(y),'actual=',xp(x));
                end;
            stage:=0;
            end;
        end; {try_proof.scan.apply_fusion.check_fusions}

	{The rules that relate to fusion are awfully complex. Basically
	 it is: (alpha, A o B) -> (beta, A)  (gamma, B), where beta
	 and gamma are formulas that must fulfill the whole plethora
	 of conditions. These are specified on page 53 of Kama Sutra.
	 It's clear that both apply_fusion and and check_fusions somehow 
	 go about checking these conditions and eventually implementing the 
	 rule Po''. G.M.}


        begin {try_proof.scan.apply_fusion}
        with tree[n] do
          begin
          if (stage=0) then check_fusions;
          tr:=reps;
          if stage<2 then tr[pcloc]:=tr[pcloc]-1;
          state:=initial;
	  if lson=0 then state:=getleft;
	  while state=initial do
		begin
		if scan(lson)<>failed
		   then begin
			if rson=0 then state:=getright;
			while state=initial do
				begin
				state:=scan(rson);
				if tree[lson].state < state
				   then state:=tree[lson].state;
				if state=failed then purge(rson);
				if rson=0 then state:=getright;
				end;
			end
		   else state:=failed;
		if state=failed then purge(lson);
		if lson=0 then state:=getleft;
		end;
          if state=failed then state:=getpc;
          end;
        end; {try_proof.scan.apply_fusion}

    begin {try_proof.scan}
    with tree[n] do
        begin

	{Translate the set of generators tree[n].gens to the array of
	 generators, gray. G.M.}

        trans(gens,gray);
        lson:=lp; rson:=rp; {left pointer and right pointer. G.M.}
        case state of
          linked  : begin {do nothing} end;

		    {Linked nodes, being identical to some other nodes
		     don't deserve any work. G.M.}

          failed  : error(3,n); {An "initial" node cannot be failed, if
				 that happened - quit. G.M.}

          initial : begin
                    count[made]:=count[made]+1;

		    {In the first place check if the formula at this node
		     is a theorem. G.M.}

                    if chk_theorem
                       then state:=theorem {If it is, then say so and quit
					    this case statement - effectively
					    at this stage we'll also quit
					    scan. G.M.}

		       {If it is not a theorem then we start the processing
			of the formula. G.M.}
					    
                       else case sym[pc] of
                         {&}  conjunction : begin
				 
				 {If the main connective is a conjunction,
				  then we proceed to apply the P& rule (see
				  page 52 of Kama Sutra). This rule states:
				  [alpha, A&B] -> alpha,A   alpha,B. 
				  The two create statements below fork two
				  new branches in agreement with P&. 
				  Remember that these statements will
				  set state of the two newly created nodes
				  to INITIAL. G.M.}

                                            create(lson,lsub[pc],0);
                                            create(rson,rsub[pc],0);

					    {Now descend recursively down the
					     left branch and if it's OK...
					     G.M.}

                                            if scan(lson)<>failed
                                               then begin

						    {... then descend 
						     recursively down the right
						     branch. G.M.}

						    state:=scan(rson);

						    {If the returned state
						     for the right branch is
						     better than the stage
						     for the left branch then
						     make the state of the
						     present node equal to
						     the state of the left
						     branch. I.e., the worse
						     wins. G.M.}

						    if tree[lson].state < state
						       then 
						       state:=tree[lson].state;
						    end
                                               else state:=failed;

					       {If any of the two branches
						failed then this node is
						also failed. G.M.}

                                            end;
                         {+}  fission     : begin {Apply P+, which is
						   exactly what the create
						   statement below does.
						   No new branch is forked.
						   G.M.}

                                            create(lson,lsub[pc],rsub[pc]);

					    {...and descend recursively down
					     this branch invoking scan again.
					     G.M.}

                                            state:=scan(lson);
                                            end;
                         {V}  disjunction,
                         {o}  fusion      : {Just do check against Curry
					     property and run filter, i.e.
					     check against Matrix property
					     and parts property. G.M.}

					    if curry_fail
                                               then state:=failed
                                               else state:=filter;

                              otherwise   : {If it's neither of the connectives
					     specified above the formula is
					     incorrect. G.M.}

					    state:=failed;
                            end;  {case}
                    end;

          open    : {Again first check if the formula at the present node
		     is a theorem. If it is, clean up and go home. G.M.}
	  
		    if chk_theorem
                       then begin state:=theorem; purge(lson); purge(rson) end
                    else

		    {If it is not a theorem, continue work. G.M.}

                    case sym[pc] of
                  {+} fission     : {No forking involved in this case.
				     Continue recursively down this branch.
				     G.M.}

				    state:=scan(lson);

                  {&} conjunction : {Descend first down the left and then
				     down the right node (if need be). G.M.}
		  
				    if scan(lson)<>failed
                                       then begin
					    state:=scan(rson);
					    if tree[lson].state < state
						then state:=tree[lson].state;
					    end
                                       else state:=failed;

                  {V} disjunction : {Apply Pv'' rule. See page 52 of Kama
				     Sutra. The Pv'' rule goes basically like
				     that: 
					[alpha, AvB] -> [beta, A]
				     or
					[alpha, AvB] -> [beta, B]
				     where
					beta = alpha (cases Ba and C)
					beta = [alpha, AvB] (case Bb)
				     Notice that no forking is involved in
				     this process. What is not quite clear
				     in this piece of code is how the
				     conditions A, Ba, Bb, and C are
				     checked against. G.M.}

				    {But first check if this node is OK. G.M.}
		  
		                    if not_baddy then
                                    begin
                                    repeat  if lson=0
                                               then begin
                                                    case stage of

                                                    0: {Do [alpha, A].  G.M.}

						       create(lson,lsub[pc],0);

                                                    1: {Do [alpha, B]. G.M.}

						       begin
                                                       create(lson,rsub[pc],0);
                                                       if reps[pcloc]>1
                                                          then stage:=3;
                                                       end;

						    2: {Do [alpha, AvB, A]. 
							G.M.}

                                                       create(lson,pc,lsub[pc]);

                                                    3: {Do [alpha, AvB, B].
							G.M.}
						    
						       create(lson,pc,rsub[pc]);
                                                    end;

						    {Prepare for the next
						     stage. G.M.}

                                                    stage:=stage+1;
                                                    end;

					    {Now descend recursively down
					     this branch. G.M.}

                                            state:=scan(lson);

					    {If this branch is bad then
					     throw it away, set the state
					     of this node back to INITIAL
					     and fetch the next possible
					     principal connective. G.M.}

                                            if state=failed
                                               then if stage<4
                                                       then begin
                                                            purge(lson);
                                                            state:=initial;
                                                            end
                                                       else 
						       state:=getpc
                                               else if state=initial
                                                       then error(4,n);
                                    until state<>initial;

				    {If at the end of this whole process
				     we still didn't make it this whole
				     formula must be bad. G.M.}

                                if state=failed then add_baddy(n);
                                end;

                      fusion  : {Again, check if the formula at this node
				 is not bad. G.M.}
		      
				if not_baddy then

				{And if it isn't, then apply the fusion
				 rule Po''. This rule is very complex, and
				 its application is delegated to a specialized
				 procedure apply_fusion. G.M.}

                                begin apply_fusion;
                                      if state=failed then add_baddy(n); end;

                    otherwise : error(5,n);
                    end; {case}
        theorem : stage:=6;
        end;  {case}

	{Now we look at the node again - at this stage it should have
	 been already thoroughly processed, so it can be either a THEOREM
	 or FAILED. If the node is FAILED chop off its branches. If it
	 is a THEOREM then if it is possibly a new one (stage <> 6) try
	 to add it to the list of theorems. Note that the branches are
	 not cut off if it is a theorem, because we will need them to
	 print the proof out. G.M.}

        case state of
          theorem : if stage<>6
                       then if lson<>0
                               then add_theorem(n);
          initial : error(6,n);
          failed  : begin purge(rson); purge(lson); end;
	  otherwise : begin end;
        end; {case}

	{Clean up for exit from scan. G.M.}

        lp:=lson; rp:=rson;
        scan:=state;
        end;
    end; {try_proof.scan}


    procedure dpthstat(n:node_ptr;dpth:integer); {try_proof.dpthstat}
        var l:node_ptr;
        begin
        if n<>0
           then with tree[n] do
                begin

		{Increment depth counter. G.M.}

                dpth:=dpth+1;
                count[depth]:=max(count[depth],dpth);

                if lp=0 {If left pointer is missing, then the branch
			 terminates here. G.M.}

                   then begin
                        count[width]:=count[width]+1;
                        if not(state in [theorem,linked])
                           then count[srch]:=count[srch]+1;
                        end;

		{Descend recursively down left and right branches here.
		 If these don't exist, dpthstat will return from recursion.
		 G.M.}

                l:=lp; dpthstat(l,dpth);
                l:=rp; dpthstat(l,dpth);
                end;
        end; {try_proof.dpthstat}

    begin {try_proof}

    {Create the first node of the proof search tree. Set up the 
     timer and counters. G.M.}

    assign(root); strtime:=clock; endtime:=wallclock; chkpt:=0;
    for cc:=scans to depth do count[cc]:=0;

    with tree[root] do
        begin
        gens:=[rootsub]; reps[1]:=1; state:=initial; pc:=rootsub;
        pcloc:=1; locmax:=1;
        repeat
            count[depth]:=0; count[width]:=0; count[srch]:=0;

	    {Now call scan which will actually analyze the formula
	     and build recursively the proof search tree. G.M.}

            state:=scan(root);

	    {Check the size of the proof search tree. G.M.}

            dpthstat(root,0);

            count[maxwdth]:=max(count[width],count[maxwdth]);
            count[maxdpth]:=max(count[depth],count[maxdpth]);
            count[scans]:=count[scans]+1;

	    {Inform the user about how things are going. G.M.}

            if ((wallclock-endtime) div 60 ) > 0
               then begin
                    if chkpt=0 then writeln;
                    chkpt:=chkpt+1;
                    write('CHECKPOINT ',xp(chkpt),'at '); inform;
                    if (chkpt mod 5)=0 then info;
                    endtime:=wallclock;
                    end;

            {Question: will scan actually modify the root node, so
	     that repeating the whole process, will do something
	     different from what it did last time? G.M.}

        until state<>open;
        end;
    endtime:=(clock-strtime) div 1000;
    writeln;
    end; {try_proof}


{procedure error(n,loc:integer); forward; G.M.}

procedure error;

{This procedure flags various errors and aborts working on a formula.
 It does not abort the program, it does not clean up the mess, with
 the exception of 'Node space exhausted' condition (memchk calls auto_init(1),
 which relinks all the available nodes into a ring).
 Called from assign and try_proof.scan only.
 G.M.}

    begin
    writeln; write('ERROR at '); inform;
    write(ht,'Node  is: '); print_mset(loc,false,0);
    write(ht,'Error is: ');
    write(ht,'Dumped at '); info;
    case n of
        1 : begin
            writeln('Node space exhausted');
            memchk;
            goto 1;
            end;
        2 : writeln('Trying to assign NULL node');
        3 : writeln('SCAN treating failed node');
        4 : writeln('SCAN caught in V-loop');
        5 : writeln('SCAN applying vacuous rule');
        6 : writeln('SCAN exiting in INITIAL state');
    end; {case}
    goto 1;
    end; {error}


begin {main}
initprocedure;
auto_init(0);
auto_init(1);
strtime:=wallclock;
lino:=0; endform:=false;
1: repeat
        if pre_tabs then try_proof;
        info;
        if tree[root].state=theorem
           then begin writeln; print_mset(root,true,0) end;

   until endform;
inform;
end. {main}
